##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
pykiso - extensible framework for (embedded) integration testing.
*****************************************************************

:module: pykiso

:synopsis: ``pykiso`` is an extensible framework for (embedded) integration testing.

.. currentmodule:: pykiso

"""

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata

# get version from package metadata to automatically set the version dunder
__version__ = metadata.version(__name__)

from . import (
    auxiliary,
    cli,
    config_parser,
    connector,
    logging_initializer,
    message,
    types,
)
from .auxiliary import AuxiliaryCommon
from .connector import CChannel, Flasher
from .exceptions import (
    AuxiliaryCreationError,
    PykisoError,
    TestCollectionError,
)
from .interfaces.dt_auxiliary import DTAuxiliaryInterface
from .interfaces.mp_auxiliary import MpAuxiliaryInterface
from .interfaces.simple_auxiliary import SimpleAuxiliaryInterface
from .interfaces.thread_auxiliary import AuxiliaryInterface
from .message import Message
from .test_coordinator import test_case, test_message_handler, test_suite
from .test_coordinator.test_case import (
    BasicTest,
    RemoteTest,
    define_test_parameters,
    retry_test_case,
)
from .test_coordinator.test_suite import (
    BasicTestSuiteSetup,
    BasicTestSuiteTeardown,
    RemoteTestSuiteSetup,
    RemoteTestSuiteTeardown,
)

logging_initializer.add_internal_log_levels()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Thread based Auxiliary Interface
********************************

:module: thread_auxiliary

:synopsis: common thread based auxiliary interface

.. currentmodule:: thread_auxiliary

.. warning :: AuxiliaryInterface will be deprecated in a few releases!

"""
import abc
import logging
import queue
import threading
import time
import warnings
from typing import List

from pykiso.auxiliary import AuxiliaryCommon

from ..exceptions import AuxiliaryCreationError
from ..logging_initializer import initialize_loggers
from ..types import MsgType

log = logging.getLogger(__name__)


# Ensure lock and queue unique reference: needed because python will do
# a copy of the created object when the unittests are called
class AuxiliaryInterface(threading.Thread, AuxiliaryCommon):
    """Defines the Interface of all thread based auxiliaries.

    Auxiliaries get configured by the Test Coordinator, get instantiated by
    the TestCases and in turn use Connectors.
    """

    def __init__(
        self,
        name: str = None,
        is_proxy_capable: bool = False,
        activate_log: List[str] = None,
        auto_start: bool = True,
    ) -> None:
        """Auxiliary initialization.

        :param name: alias of the auxiliary instance
        :param is_proxy_capable: notify if the current auxiliary could
            be (or not) associated to a proxy-auxiliary.
        :param activate_log: loggers to deactivate
        :param auto_start: determine if the auxiliayry is automatically
             started (magic import) or manually (by user)
        """
        # Initialize thread class
        super().__init__()
        initialize_loggers(activate_log)
        # Save the name
        self.name = name
        # Define thread control attributes & methods
        self.lock = threading.RLock()
        self.queue_in = queue.Queue()
        self.queue_out = queue.Queue()
        self.stop_event = threading.Event()
        self.is_proxy_capable = is_proxy_capable
        # Create state
        self.is_instance = False
        self.auto_start = auto_start
        self._aux_copy = None

        warnings.warn(
            "AuxiliaryInterface will be deprecated in a few releases!",
            category=FutureWarning,
        )

    def start(self) -> None:
        """Start the thread and create the auxiliary only if auto_start
        flag is False.
        """
        if not self.is_alive():
            super().start()
            if not self.auto_start:
                self.create_instance()

    def create_instance(self) -> bool:
        """Create an auxiliary instance and ensure the communication to it.

        :return: message.Message() - Contain received message

        :raises AuxiliaryCreationError: if instance creation failed
        """
        if self.lock.acquire():
            # Trigger the internal requests
            self.queue_in.put("create_auxiliary_instance")
            # Wait until the request was processed
            report = self.queue_out.get()
            # Release the above lock
            self.lock.release()
            # aux instance can't be created just exit
            if not report:
                raise AuxiliaryCreationError(self.name)
            # Return the report
            return report

    def delete_instance(self) -> bool:
        """Delete an auxiliary instance and its communication to it.

        :return: message.Message() - Contain received message
        """
        if self.lock.acquire():
            # Trigger the internal requests
            self.queue_in.put("delete_auxiliary_instance")
            # Wait until the request was processed
            report = self.queue_out.get()
            # Release the above lock
            self.lock.release()
            # Return the report
            return report

    def run(self) -> None:
        """Run function of the auxiliary thread."""
        while not self.stop_event.is_set():
            # Step 1: Check if a request is available & process it
            request = ""
            # Check if a request was received
            if not self.queue_in.empty():
                request = self.queue_in.get_nowait()
            # Process the request
            if request == "create_auxiliary_instance" and not self.is_instance:
                # Call the internal instance creation method
                return_code = self._create_auxiliary_instance()
                # Based on the result set status:
                self.is_instance = return_code
                # Enqueue the result for the request caller
                self.queue_out.put(return_code)
            elif request == "delete_auxiliary_instance" and self.is_instance:
                # Call the internal instance delete method
                return_code = self._delete_auxiliary_instance()
                # Based on the result set status:
                self.is_instance = not return_code
                # Enqueue the result for the request caller
                self.queue_out.put(return_code)
            elif (
                isinstance(request, tuple)
                and self.is_instance
                and request[0] == "command"
            ):
                # If the instance is created, send the requested command
                # to the instance via the internal method
                _, cmd, data = request
                cmd_response = self._run_command(cmd, data)
                if cmd_response is not None:
                    self.queue_out.put(cmd_response)
            elif request == "abort" and self.is_instance:
                self.queue_out.put(self._abort_command())
            elif request != "":
                # A request was received but could not be processed
                log.internal_warning(
                    f"Unknown request '{request}', will not be processed!"
                )
                log.internal_warning(f"Aux status: {self.__dict__}")

            # Step 2: Check if something was received from the aux instance if instance was created
            if self.is_instance:
                received_message = self._receive_message(timeout_in_s=0)
                # If yes, send it via the out queue
                if received_message is not None:
                    self.queue_out.put(received_message)

            # Free up cpu usage when auxiliary is suspended
            if not self.is_instance:
                time.sleep(0.050)

        # Thread stop command was set
        log.internal_info("{} was stopped".format(self))
        # Delete auxiliary external instance if not done
        if self.is_instance:
            self._delete_auxiliary_instance()

    @abc.abstractmethod
    def _create_auxiliary_instance(self) -> bool:
        """Create the auxiliary instance with witch we will communicate.

        :return: True - Successfully created / False - Failed by creation

        .. note: Errors should be logged via the logging with the right level
        """
        pass

    @abc.abstractmethod
    def _delete_auxiliary_instance(self) -> bool:
        """Delete the auxiliary instance with witch we will communicate.

        :return: True - Successfully deleted / False - Failed deleting

        .. note: Errors should be logged via the logging with the right level
        """
        pass

    @abc.abstractmethod
    def _run_command(self, cmd_message: MsgType, cmd_data: bytes = None) -> MsgType:
        """Run a command for the auxiliary.

        :param cmd_message: command in form of a message to run

        :param cmd_data: payload data for the command

        :return: True - Successfully received by the instance / False - Failed sending

        .. note: Errors should be logged via the logging with the right level
        """
        pass

    @abc.abstractmethod
    def _abort_command(self) -> bool:
        """Abort the sent command."""
        pass

    @abc.abstractmethod
    def _receive_message(self, timeout_in_s: float) -> MsgType:
        """Defines what needs to be done as a receive message. Such as,
            what do I need to do to receive a message.

        :param timeout_in_s: How much time to block on the receive

        :return: message.Message - If one received / None - Else
        """
        pass



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Multiprocessing Proxy Channel
*****************************

:module: cc_mp_proxy

:synopsis: concrete implementation of a multiprocessing proxy channel

CCProxy channel was created, in order to enable the connection of
multiple auxiliaries on one and only one CChannel. This CChannel
has to be used with a so called proxy auxiliary.

.. currentmodule:: cc_mp_proxy

"""
from __future__ import annotations

import logging
import multiprocessing
import queue
from typing import TYPE_CHECKING, Any

from pykiso.connector import CChannel

if TYPE_CHECKING:
    from pykiso.lib.auxiliaries.mp_proxy_auxiliary import MpProxyAuxiliary
    from pykiso.types import ProxyReturn

log = logging.getLogger(__name__)


class CCMpProxy(CChannel):
    """Multiprocessing Proxy CChannel for multi auxiliary usage."""

    def __init__(self, **kwargs):
        """Initialize attributes."""
        kwargs.update(processing=True)
        super().__init__(**kwargs)
        # instantiate directly both queue_in and queue_out
        self.queue_in = multiprocessing.Queue()
        self.queue_out = multiprocessing.Queue()
        self.timeout = 1
        # used for physical channel attributes access from main auxiliary
        self._proxy = None
        self._physical_channel = None

    def _bind_channel_info(self, proxy_aux: MpProxyAuxiliary):
        """Bind a :py:class:`~pykiso.lib.auxiliaries.mp_proxy_auxiliary.MpProxyAuxiliary`
        instance that is instanciated in order to handle the connection of
        multiple auxiliaries to a single communication channel in order to
        hide the underlying proxy setup.

        :param proxy_aux: the proxy auxiliary instance that is holding the
            real communication channel.
        """
        self._proxy = proxy_aux
        self._physical_channel = proxy_aux.channel

    def __getattr__(self, name: str) -> Any:
        """Implement getattr to retrieve attributes from the real channel attached
        to the underlying :py:class:`~pykiso.lib.auxiliaries.mp_proxy_auxiliary.MpProxyAuxiliary`.

        :param name: name of the attribute to get.
        :raises AttributeError: if the attribute is not part of the real
            channel instance or if the real channel hasn't been bound to
            this proxy channel yet.
        :return: the found attribute value.
        """
        if self._physical_channel is not None:
            with self._proxy.lock:
                return getattr(self._physical_channel, name)
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute '{name}'"
        )

    def __getstate__(self):
        """Avoid getattr to be called on pickling before instanciation,
        which would cause infinite recursion.
        """
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def _cc_open(self) -> None:
        """Open proxy channel.

        Due to usage of multiprocessing the queue_in and queue_out state
        doesn't have to change in order to ensure that ProxyAuxiliary
        works even if suspend or resume is called.
        """
        log.internal_debug("Open proxy channel")

    def _cc_close(self) -> None:
        """Close proxy channel.

        Due to usage of multiprocessing the queue_in and queue_out state
        doesn't have to change in order to ensure that ProxyAuxiliary
        works even if suspend or resume is called.
        """
        log.internal_debug("Close proxy channel")

    def _cc_send(self, *args: tuple, **kwargs: dict) -> None:
        """Populate the queue in of the proxy connector.

        :param args: tuple containing positionnal arguments
        :param kwargs: dictionary containing named arguments
        """
        log.internal_debug(f"put at proxy level: {args} {kwargs}")
        self.queue_in.put((args, kwargs))

    def _cc_receive(self, timeout: float = 0.1) -> ProxyReturn:
        """Depopulate the queue out of the proxy connector.

        :param timeout: not used

        :return: bytes and source when it exist. if queue timeout
            is reached return None
        """
        try:
            return_response = self.queue_out.get(True, self.timeout)
            log.internal_debug(f"received at proxy level : {return_response}")
            return return_response
        except queue.Empty:
            return {"msg": None}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Simple Auxiliary Interface
**************************

:module: simple_auxiliary

:synopsis: common auxiliary interface for very simple auxiliary
    (without usage of thread or multiprocessing)

.. currentmodule:: simple_auxiliary

"""
import abc
import logging
from typing import List, Optional

from ..exceptions import AuxiliaryCreationError
from ..logging_initializer import add_logging_level, initialize_loggers
from .thread_auxiliary import AuxiliaryInterface

log = logging.getLogger(__name__)


class SimpleAuxiliaryInterface(metaclass=abc.ABCMeta):
    """Define the interface for all simple auxiliary where usage of
    thread or mulitprocessing is not necessary.
    """

    def __new__(cls, *args, **kwargs):
        """Create instance and add internal kiso log levels in
        case the auxiliary is used outside the pykiso context
        """
        if not hasattr(logging, "INTERNAL_WARNING"):
            add_logging_level("INTERNAL_WARNING", logging.WARNING + 1)
            add_logging_level("INTERNAL_INFO", logging.INFO + 1)
            add_logging_level("INTERNAL_DEBUG", logging.DEBUG + 1)
        return super(SimpleAuxiliaryInterface, cls).__new__(cls)

    def __init__(self, name: str = None, activate_log: List[str] = None) -> None:
        """Auxiliary initialization.

        :param activate_log: loggers to deactivate
        :param name: alias of the auxiliary instance
        """
        self.name = name
        self.is_instance = False
        initialize_loggers(activate_log)

    def __repr__(self) -> str:
        name = self.name if self.name is not None else ""
        repr_ = super().__repr__()
        if name:
            repr_ = repr_[:1] + f"{name} is " + repr_[1:]
        return repr_

    def create_instance(self) -> bool:
        """Create an auxiliary instance and ensure the communication to it.

        :return: True if creation was successful otherwise False

        :raises AuxiliaryCreationError: if instance creation failed
        """
        self.is_instance = self._create_auxiliary_instance()
        if not self.is_instance:
            raise AuxiliaryCreationError(self.name)
        return self.is_instance

    def delete_instance(self) -> bool:
        """Delete an auxiliary instance and its communication to it.

        :return: True if deletion was successful otherwise False
        """
        report = self._delete_auxiliary_instance()
        self.is_instance = not report
        return report

    def resume(self) -> None:
        """Resume current auxiliary's run, by running the
        create_instance method in the background.

        .. warning:: due to the usage of create_instance if an issue
            occurred the exception AuxiliaryCreationError is raised.
        """
        if not self.is_instance:
            self.create_instance()
        else:
            log.internal_warning(f"Auxiliary '{self}' is already running")

    def suspend(self) -> None:
        """Suspend current auxiliary's run."""
        if self.is_instance:
            self.delete_instance()
        else:
            log.internal_warning(f"Auxiliary '{self}' is already stopped")

    def stop(self):
        """Stop the auxiliary"""
        self._delete_auxiliary_instance()

    @abc.abstractmethod
    def _create_auxiliary_instance(self) -> bool:
        """Create the auxiliary instance with which we will communicate.

        :return: True - Successfully created / False - Failed by creation

        .. note: Errors should be logged via the logging with the right level
        """
        pass

    @abc.abstractmethod
    def _delete_auxiliary_instance(self) -> bool:
        """Delete the auxiliary instance with witch we will communicate.

        :return: True - Successfully deleted / False - Failed deleting

        .. note: Errors should be logged via the logging with the right level
        """
        pass



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via segger j-link
***************************************

:module: cc_rtt_segger

:synopsis: channel used to enable RTT communication using Segger J-Link debugger.
    Additionally, RTT logs can be captured by setting the rtt_log_path parameter
    on the specified channel.

.. currentmodule:: cc_rtt_segger

"""
import functools
import logging
import threading
import time
from pathlib import Path
from typing import Dict, Optional, Union

import pylink

from pykiso import connector
from pykiso.message import Message

log = logging.getLogger(__name__)


def _need_connection(f):
    """Decorator to check the JLink is connected and raises an error otherwise"""

    @functools.wraps(f)
    def check_before_execution(self, *args, **kwargs):
        """Check if Jlink is opened, else raise error

        :raises pylink.JLinkException: if Jlink not available
        """
        if not self.jlink.opened():
            raise pylink.JLinkException("You need to connect to a JLink first")
        return f(self, *args, **kwargs)

    return check_before_execution


def _need_rtt(f):
    """Decorator to check that a RTT connection has been configured and raises an error otherwise"""

    @functools.wraps(f)
    def check_before_execution(self, *args, **kwargs):
        """Check if Jlink is opened and RTT configured, else raise errors

        :raises pylink.JLinkException: if RTT config not done or Jlink not available
        """

        if not self.jlink.opened():
            raise pylink.JLinkException("You need to connect to a JLink first")
        if not self.rtt_configured:
            raise pylink.JLinkException("You need to configure RTT first")
        return f(self, *args, **kwargs)

    return check_before_execution


class CCRttSegger(connector.CChannel):
    """Channel using RTT to communicate through Segger J-Link debugger."""

    def __init__(
        self,
        serial_number: int = None,
        chip_name: str = "STM32L562QE",
        speed: int = 4000,
        block_address: int = 0x2003F800,
        verbose: bool = False,
        tx_buffer_idx: int = 3,
        rx_buffer_idx: int = 0,
        rtt_log_path: Optional[str] = None,
        rtt_log_buffer_idx: int = 0,
        rtt_log_speed: float = 1000,
        connection_timeout: int = 5,
        **kwargs,
    ):
        """Initialize attributes.

        :param serial_number: optional segger debugger serial number (required if many connected)
        :param chip_name: microcontoller name (STM....)
        :param speed: communication speed in Hz
        :param block_address: start address to start RTT communication
        :param tx_buffer_idx: buffer index used for transmission
        :param rx_buffer_idx: buffer index used for reception
        :param verbose: boolean indicating if J-Link connection should be verbose in logging
        :param rtt_log_path: path to the folder where the RTT log file should be stored
        :param rtt_log_buffer_idx: buffer index used for RTT logging
        :param rtt_log_speed: number of log per second to be pulled (manage the CPU load for logging)
            None value fetch log at the CPU's speed. Default 1000 logs/s
        :param connection_timeout: available time (in seconds) to open the connection
        """
        super().__init__(**kwargs)
        self.serial_number = serial_number if isinstance(serial_number, int) else None
        self.chip_name = chip_name
        self.speed = speed
        self.block_address = block_address
        self.tx_buffer_idx = tx_buffer_idx
        self.rx_buffer_idx = rx_buffer_idx
        self.verbose = verbose
        self.jlink = None
        self.connection_timeout = connection_timeout
        self.rtt_log_buffer_idx = rtt_log_buffer_idx
        self.rx_buffer_size = None
        # initialize rtt logging specific parameters
        self.rtt_configured = False
        self._log_thread_running = False
        self.rtt_log_refresh_time = round(1 / rtt_log_speed, 6) if rtt_log_speed else 0
        self.rtt_log_thread = threading.Thread(target=self.receive_log)
        self.rtt_log_path = rtt_log_path
        self.rtt_log = logging.getLogger(f"{__name__}{serial_number or ''}.RTT")
        if self.rtt_log_path is not None:
            self.rtt_log_buffer_size = 0
            self.rtt_log_path = Path(rtt_log_path)
            # if log folder does not exists create it during process
            self.rtt_log_path.mkdir(parents=True, exist_ok=True)
            rtt_fh = logging.FileHandler(self.rtt_log_path / "rtt.log")
            rtt_fh.setLevel(logging.DEBUG)
            self.rtt_log.setLevel(logging.DEBUG)
            self.rtt_log.addHandler(rtt_fh)

    def read_target_memory(
        self, addr: int, num_units: int, zone: str = None, nbits: int = 32
    ) -> Optional[list]:
        """Read the given target's memory units and the given address.

        .. note:: The optional ``zone`` specifies a memory zone to
            access to read from, e.g. ``IDATA``, ``DDATA``, or ``CODE``.

        .. warning:: The given number of bits, if provided, must be
            either ``8``, ``16``, or ``32``.  If not provided, always
            reads 32 bits.

        :param addr: start address to read from
        :param num_units: number of units to read
        :param zone: optional memory zone name to access
        :param nbits: number of bits to use for each unit

        :return: List of units read from the target.
        """
        mem_vals = None

        try:
            mem_vals = self.jlink.memory_read(addr, num_units, zone, nbits)
        except pylink.errors.JLinkException:
            log.exception(f"encountered error while reading memory at {hex(addr)}")
        except ValueError:
            log.exception("wrong number of bits given : must be 8, 16 or 32 bits")

        return mem_vals

    def _cc_open(self) -> None:
        """Connect debugger/microcontroller.

        This method proceed to the following actions :
        - create a JLink class instance
        - connect to  the debugger(using open method)
        - set debugger interface to SWD
        - connect debugger to the specified chip
        - start RTT communication
        - start RTT Logging the specified channel if activated

        :raise JLinkRTTException: if connection timeout occurred.
        """
        self.jlink = pylink.JLink()

        # connect to J-Link debugger
        if not self.jlink.opened():
            self.jlink.open(self.serial_number)
            log.internal_info(
                f"connection made with J-Link debugger {self.serial_number}"
            )
        else:
            log.internal_debug("connection to J-Link already started")
        # set target interface to SWD
        self.jlink.set_tif(pylink.enums.JLinkInterfaces.SWD)
        # connect debugger to  the specified target
        self.jlink.connect(
            chip_name=self.chip_name, speed=self.speed, verbose=self.verbose
        )
        log.internal_debug(
            f"connection to chip {self.chip_name} performed at speed {self.speed}Hz"
        )
        # reset debugger if halted
        if self.jlink.halted():
            log.internal_info(
                f"J-Link is halted, reset target and wait for {self.connection_timeout}s"
            )
            self.jlink.reset(halt=False)
            time.sleep(self.connection_timeout)
        # start rtt at the specified address
        self.jlink.rtt_start(self.block_address)
        log.internal_info(f"RTT communication started at address {self.block_address}")

        t_start = time.perf_counter()
        while True:
            try:
                num_up = self.jlink.rtt_get_num_up_buffers()
                num_down = self.jlink.rtt_get_num_down_buffers()
                log.internal_debug(
                    f"RTT started. Found {num_up} up and {num_down} down channels."
                )
                # get rx buffer size
                rx_buffer = self.jlink.rtt_get_buf_descriptor(self.rx_buffer_idx, True)
                self.rx_buffer_size = rx_buffer.SizeOfBuffer
                log.internal_debug(
                    f"Maximum size for a received message set to {self.rx_buffer_size}"
                )
                self.rtt_configured = True
                break
            except pylink.errors.JLinkRTTException:
                time.sleep(0.1)
                # Exit the while loop once timeout is reached
                if time.perf_counter() > (t_start + self.connection_timeout):
                    raise

        # start rtt logging thread on buffer index rtt_log_buffer_idx
        if self.rtt_log_path is not None:
            try:
                rtt_log_buffer = self.jlink.rtt_get_buf_descriptor(
                    self.rtt_log_buffer_idx, True
                )
                self.rtt_log_buffer_size = rtt_log_buffer.SizeOfBuffer
                if self.rtt_log_buffer_size == 0:
                    raise ValueError
                log.internal_debug(
                    f"RTT log buffer size is {self.rtt_log_buffer_size} bytes"
                )
            except ValueError:
                log.internal_debug("Read RTT log buffer size is 0, defaulting to 1kB")
                self.rtt_log_buffer_size = 1024
            except pylink.errors.JLinkRTTException as e:
                log.error(f"Could not get RTT log buffer size: {e}")
                if self.rtt_log_buffer_idx not in range(num_up + 1):
                    log.error(f"Buffer index must be at most {num_up}")
                    self.rtt_log_buffer_idx = 0
                self.rtt_log_buffer_size = 1024
            finally:
                self._log_thread_running = True
                self.rtt_log_thread.start()
                log.internal_info("RTT logging started")

    def _cc_close(self) -> None:
        """Close current RTT communication in use."""

        if self.jlink is not None:
            if self._log_thread_running:
                self._log_thread_running = False
                self.rtt_log_thread.join()
            self.jlink.rtt_stop()
            self.jlink.close()
            log.internal_info("RTT communication closed")

    def _cc_send(self, msg: bytes) -> None:
        """Send message using the corresponding RTT buffer.

        :param msg: message to send, should be bytes.
        """
        try:

            msg = list(msg)
            bytes_written = self.jlink.rtt_write(self.tx_buffer_idx, msg)
            log.internal_debug(
                "===> message sent (RTT) on buffer %d: %s, number of bytes written: %d",
                self.tx_buffer_idx,
                msg,
                bytes_written,
            )
        except Exception:
            log.exception(
                f"ERROR occurred while sending {len(msg)} bytes on buffer {self.tx_buffer_idx}"
            )

    def _cc_receive(
        self, timeout: float = 0.1, size: int = None, **kwargs
    ) -> Dict[str, Optional[bytes]]:
        """Read message from the corresponding RTT buffer.

        :param timeout: timeout applied on receive event
        :param size: maximum amount of bytes to read
        :return: dictionary containing the received bytes if successful, otherwise None
        """
        is_timeout = False
        # maximum amount of bytes to read out
        size = size or self.rx_buffer_size
        t_start = time.perf_counter()

        # rtt_read is not a blocking method due to this fact a while loop is used
        # to act like a blocking ones.
        while not is_timeout:
            try:
                # Read the message header or all of the buffer
                msg_received = self.jlink.rtt_read(self.rx_buffer_idx, size)
                # if a message is received
                if msg_received:
                    # Parse the bytes list into bytes string
                    msg_received = bytes(msg_received)
                    log.internal_debug(
                        "<=== message received (RTT) on buffer %d: %s, number of bytes read: %d",
                        self.rx_buffer_idx,
                        msg_received,
                        len(msg_received),
                    )
                    break
            except Exception:
                log.exception(
                    f"encountered error while receiving message via {self} on buffer {self.rx_buffer_idx}"
                )
                return {"msg": None}

            # Exit the while loop once timeout is reached
            if time.perf_counter() > (t_start + timeout):
                is_timeout = True
                msg_received = None

        return {"msg": msg_received}

    @_need_rtt
    def receive_log(self) -> None:
        """Receive RTT log messages from the corresponding RTT buffer."""
        while self._log_thread_running:
            # receive at most rtt_log_buffer_size of RTT logs
            log_msg = self.jlink.rtt_read(
                self.rtt_log_buffer_idx, self.rtt_log_buffer_size
            )
            if log_msg:
                self.rtt_log.internal_debug(bytes(log_msg).decode())
            time.sleep(self.rtt_log_refresh_time)  # reduce resource consumption

    @_need_connection
    def reset_target(self, wait_time: int = 100, halt: bool = False) -> None:
        """Reset target via JLink.

        :param wait_time: Amount of milliseconds to delay after reset
        :param halt: if the CPU should halt after reset
        """
        logging.info("Reset Target")
        self.jlink.enable_reset_pulls_reset()
        self.jlink.reset(wait_time, halt)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
CommunicationAuxiliary
**********************

:module: communication_auxiliary

:synopsis: Auxiliary used to send raw bytes via a connector instead of
    pykiso.Messages

.. currentmodule:: communication_auxiliary


"""
from __future__ import annotations

import functools
import logging
import queue
import threading
from contextlib import ContextDecorator
from typing import Any, Optional

from pykiso import CChannel, Message
from pykiso.interfaces.dt_auxiliary import (
    DTAuxiliaryInterface,
    close_connector,
    open_connector,
)

log = logging.getLogger(__name__)


class _collect_messages(ContextDecorator):
    """Context manager and decorator for the communication auxiliary
    allowing messages collection (putting them in the queue).
    """

    def __init__(self, com_aux: CommunicationAuxiliary):
        """Constructor used to inherit some of communication auxiliary
        features.
        """
        self.com_aux = com_aux

    def __enter__(self):
        """Set the queue event to allow messages collection."""
        log.internal_debug("Start queueing received messages.")
        self.com_aux.queueing_event.set()

    def __exit__(self, *exc):
        """Clear queue event to stop messages collection."""
        log.internal_debug("Stop queueing received messages.")
        self.com_aux.queueing_event.clear()


class CommunicationAuxiliary(DTAuxiliaryInterface):
    """Auxiliary used to send raw bytes via a connector instead of pykiso.Messages."""

    def __init__(self, com: CChannel, **kwargs: dict) -> None:
        """Constructor.

        :param com: CChannel that supports raw communication
        """
        super().__init__(
            is_proxy_capable=True, tx_task_on=True, rx_task_on=True, **kwargs
        )
        self.channel = com
        self.queue_tx = queue.Queue()
        self.queueing_event = threading.Event()
        self.collect_messages = functools.partial(_collect_messages, com_aux=self)

    @open_connector
    def _create_auxiliary_instance(self) -> bool:
        """Open the connector communication.

        :return: True if the channel is correctly opened otherwise False
        """
        log.internal_info("Auxiliary instance created")
        return True

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """
        log.internal_info("Auxiliary instance deleted")
        return True

    def send_message(self, raw_msg: bytes) -> bool:
        """Send a raw message (bytes) via the communication channel.

        :param raw_msg: message to send

        :return: True if command was executed otherwise False
        """
        return self.run_command("send", raw_msg)

    def run_command(
        self,
        cmd_message: Any,
        cmd_data: Any = None,
        blocking: bool = True,
        timeout_in_s: int = None,
    ) -> bool:
        """Send a request by transmitting it through queue_in and
        populate queue_tx with the command verdict (successful or not).

        :param cmd_message: command to send
        :param cmd_data: data you would like to populate the command
            with
        :param blocking: If you want the command request to be
            blocking or not
        :param timeout_in_s: Number of time (in s) you want to wait
            for an answer

        :return: True if the request is correctly executed otherwise
            False
        """
        with self.lock:
            log.internal_debug(
                f"sending command '{cmd_message}' with payload {cmd_data} using {self.name} aux."
            )
            state = None
            self.queue_in.put((cmd_message, cmd_data))
            try:
                state = self.queue_tx.get(blocking, timeout_in_s)
                log.internal_debug(
                    f"command '{cmd_message}' successfully sent for {self.name} aux"
                )
            except queue.Empty:
                log.error(
                    f"no feedback received regarding request {cmd_message} for {self.name} aux."
                )
        return state

    def receive_message(
        self,
        blocking: bool = True,
        timeout_in_s: float = None,
    ) -> Optional[bytes]:
        """Receive a raw message.

        :param blocking: wait for message till timeout elapses?
        :param timeout_in_s: maximum time in second to wait for a response

        :returns: raw message
        """

        # Evaluate if we are in the context manager or not
        in_ctx_manager = False
        if self.queueing_event.is_set():
            in_ctx_manager = True

        log.internal_debug(
            f"retrieving message in {self} (blocking={blocking}, timeout={timeout_in_s})"
        )
        # In case we are not in the context manager, we have a enable the receiver thread (and afterwards disable it)
        if not in_ctx_manager:
            self.queueing_event.set()
        response = self.wait_for_queue_out(blocking=blocking, timeout_in_s=timeout_in_s)
        if not in_ctx_manager:
            self.queueing_event.clear()

        log.internal_debug(f"retrieved message '{response}' in {self}")

        # if queue.Empty exception is raised None is returned so just
        # directly return it
        if response is None:
            return None

        msg = response.get("msg")
        remote_id = response.get("remote_id")

        # stay with the old return type to not making a breaking change
        if remote_id is not None:
            return (msg, remote_id)
        return msg

    def clear_buffer(self) -> None:
        """Clear buffer from old stacked objects"""
        log.internal_info("Clearing buffer. Previous responses will be deleted.")
        with self.queue_out.mutex:
            self.queue_out.queue.clear()

    def _run_command(self, cmd_message: str, cmd_data: bytes = None) -> bool:
        """Run the corresponding command.

        :param cmd_message: command type
        :param cmd_data: payload data to send over CChannel

        :return: True if command is executed otherwise False
        """
        state = False
        if cmd_message == "send":
            try:
                self.channel.cc_send(msg=cmd_data)
                state = True
            except Exception:
                log.exception(
                    f"encountered error while sending message '{cmd_data}' to {self.channel}"
                )
        elif isinstance(cmd_message, Message):
            log.internal_debug(f"ignored command '{cmd_message} in {self}'")
        else:
            log.internal_warning(f"received unknown command '{cmd_message} in {self}'")

        self.queue_tx.put(state)

    def _receive_message(self, timeout_in_s: float) -> None:
        """Get a message from the associated channel. And put the message in
        the queue, if threading event is set.

        :param timeout_in_s: maximum amount of time (seconds) to wait
            for a message
        """
        try:
            rcv_data = self.channel.cc_receive(timeout=timeout_in_s)
            log.internal_debug(f"received message '{rcv_data}' from {self.channel}")
            msg = rcv_data.get("msg")
            if msg is not None and self.queueing_event.is_set():
                self.queue_out.put(rcv_data)
        except Exception:
            log.exception(
                f"encountered error while receiving message via {self.channel}"
            )



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Can Communication Channel using PCAN hardware
*********************************************

:module: cc_pcan_can

:synopsis: CChannel implementation for CAN(fd) using PCAN API from python-can

.. currentmodule:: cc_pcan_can

"""

import logging
import sys
from pathlib import Path
from typing import Dict, Optional, Union

import can
import can.bus
import can.interfaces.pcan.basic as PCANBasic

from pykiso import CChannel, Message

MessageType = Union[Message, bytes]

log = logging.getLogger(__name__)


class PcanFilter(logging.Filter):
    """Filter specific pcan logging messages"""

    def filter(self, record: logging.LogRecord) -> bool:
        """Determine if the specified record is to be logged. It will not if
        it is a pcan bus error message

        :param record: record of the event to filter if it is a pcan bus error

        :return: True if the record should be logged, or False otherwise.
        """
        return not record.getMessage().startswith("Bus error: an error counter")


class CCPCanCan(CChannel):
    """CAN FD channel-adapter."""

    def __init__(
        self,
        interface: str = "pcan",
        channel: str = "PCAN_USBBUS1",
        state: str = "ACTIVE",
        trace_path: str = "",
        trace_size: int = 10,
        bitrate: int = 500000,
        is_fd: bool = True,
        enable_brs: bool = False,
        f_clock_mhz: int = 80,
        nom_brp: int = 2,
        nom_tseg1: int = 63,
        nom_tseg2: int = 16,
        nom_sjw: int = 16,
        data_brp: int = 4,
        data_tseg1: int = 7,
        data_tseg2: int = 2,
        data_sjw: int = 2,
        is_extended_id: bool = False,
        remote_id: int = None,
        can_filters: list = None,
        logging_activated: bool = True,
        bus_error_warning_filter: bool = False,
        **kwargs,
    ):
        """Initialize can channel settings.

        :param interface: python-can interface modules used
        :param channel: the can interface name
        :param state: BusState of the channel
        :param trace_path: path to write the trace
        :param trace_size: maximum size of the trace (in MB)
        :param bitrate: Bitrate of channel in bit/s,ignored if using CanFD
        :param is_fd: Should the Bus be initialized in CAN-FD mode
        :param enable_brs: sets the bitrate_switch flag to use higher transmission speed
        :param f_clock_mhz:  Clock rate in MHz
        :param nom_brp: Clock prescaler for nominal time quantum
        :param nom_tseg1: Time segment 1 for nominal bit rate, that is,
            the number of quanta from the Sync Segment to the sampling point
        :param nom_tseg2: Time segment 2 for nominal bit rate,
            that is, the number of quanta from the sampling point to the end of the bit
        :param nom_sjw: Synchronization Jump Width for nominal bit rate.
            Decides the maximum number of time quanta that the controller
            can resynchronize every bit
        :param data_brp: Clock prescaler for fast data time quantum
        :param data_tseg1: Time segment 1 for fast data bit rate, that is,
            the number of quanta from the Sync Segment to the sampling point
        :param data_tseg2: Time segment 2 for fast data bit rate, that is,
            the number of quanta from the sampling point to the end of the bit.
            In the range (1..16)
        :param data_sjw: Synchronization Jump Width for fast data bit rate
        :param is_extended_id: This flag controls the size of the arbitration_id field
        :param remote_id: id used for transmission
        :param can_filters: iterable used to filter can id on reception
        :param logging_activated: boolean used to disable logfile creation
        :param bus_error_warning_filter : if True filter the logging message
        ('Bus error: an error counter')
        """
        super().__init__(**kwargs)
        self.interface = interface
        self.channel = channel
        self.state = can.bus.BusState[state.upper()]
        self.trace_path = Path(trace_path)
        self.trace_size = trace_size
        self.bitrate = bitrate
        self.is_fd = is_fd
        self.enable_brs = enable_brs
        self.f_clock_mhz = f_clock_mhz
        self.nom_brp = nom_brp
        self.nom_tseg1 = nom_tseg1
        self.nom_tseg2 = nom_tseg2
        self.nom_sjw = nom_sjw
        self.data_brp = data_brp
        self.data_tseg1 = data_tseg1
        self.data_tseg2 = data_tseg2
        self.data_sjw = data_sjw
        self.is_extended_id = is_extended_id
        self.remote_id = remote_id
        self.can_filters = can_filters
        self.bus = None
        self.logging_activated = logging_activated
        self.raw_pcan_interface = None
        self.logging_path = None
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6
        """
        Extract the base logging directory from the logging module, so we can
        create our logging folder in the correct place.
        logging_path will be set to the parent directory of the logfile which
        is set in the logging module + /raw/PCAN
        If an AttributeError occurs, the logging module is not set to log into
        a file.
        In this case we set the logging_path to None and will just log into a
        generic logfile in the current working directory, which will be
        overwritten every time, a log is initiated.
        """
        if self.trace_path.is_file():
            self.trace_path = self.trace_path.parent
            log.internal_warning(
                f"File names are not supported for trace file creation. Trace will be written to {self.trace_path}"
            )

        if not 0 < self.trace_size <= 100:
            self.trace_size = 10
            log.internal_warning(
                f"Make sure trace size is between 1 and 100 Mb. Setting trace size to default value "
                f"value : {self.trace_size}."
            )

        if bus_error_warning_filter:
            logging.getLogger("can.pcan").addFilter(PcanFilter())

        if self.enable_brs and not self.is_fd:
            log.internal_warning(
                "Bitrate switch will have no effect because option is_fd is set to false."
            )

    def _cc_open(self) -> None:
        """Open a can bus channel, set filters for reception and activate PCAN log."""
        self.bus = can.interface.Bus(
            interface=self.interface,
            channel=self.channel,
            state=self.state,
            bitrate=self.bitrate,
            fd=self.is_fd,
            f_clock_mhz=self.f_clock_mhz,
            nom_brp=self.nom_brp,
            nom_tseg1=self.nom_tseg1,
            nom_tseg2=self.nom_tseg2,
            nom_sjw=self.nom_sjw,
            data_brp=self.data_brp,
            data_tseg1=self.data_tseg1,
            data_tseg2=self.data_tseg2,
            data_sjw=self.data_sjw,
            can_filters=self.can_filters,
        )

        if self.logging_activated and self.raw_pcan_interface is None:
            self.raw_pcan_interface = PCANBasic.PCANBasic()
            self._pcan_configure_trace()

    def _pcan_configure_trace(self) -> None:
        """Configure PCAN dongle to create a trace file.

        If self.logging_path is set, this path will be created, if it does not
        exist and the logfile will be placed there.
        Otherwise it will be logged to the current working directory if a
        default filename, which will be overwritten in successive calls.
        If an error occurs, the trace will not be started and the error logged.
        No exception is thrown in this case.
        """
        pcan_channel = getattr(PCANBasic, self.channel)
        if self.trace_path is None:
            log.internal_warning(
                "No trace path specified, an existing trace will be overwritten."
            )
            pcan_path_argument = PCANBasic.TRACE_FILE_OVERWRITE
        else:
            pcan_path_argument = PCANBasic.TRACE_FILE_DATE | PCANBasic.TRACE_FILE_TIME
        try:
            if self.trace_path is not None:
                if not Path(self.trace_path).exists():
                    Path(self.trace_path).mkdir(parents=True, exist_ok=True)
                    log.internal_info(f"Path {self.trace_path} created")
                self._pcan_set_value(
                    pcan_channel,
                    PCANBasic.PCAN_TRACE_LOCATION,
                    bytes(self.trace_path),
                )
                log.internal_info(
                    f"Tracefile path in PCAN device configured to {self.trace_path}"
                )

            if sys.platform != "darwin":
                log.internal_info("Segmented option of trace file activated.")
                self._pcan_set_value(
                    pcan_channel,
                    PCANBasic.TRACE_FILE_SEGMENTED,
                    PCANBasic.PCAN_PARAMETER_ON,
                )

                if self.trace_size != 10:
                    log.internal_info(f"Trace size set to {self.trace_size} MB.")
                    self._pcan_set_value(
                        pcan_channel, PCANBasic.PCAN_TRACE_SIZE, self.trace_size
                    )
            else:
                log.internal_warning("TRACE_FILE_SEGMENTED deactivated for macos!")

            self._pcan_set_value(
                pcan_channel,
                PCANBasic.PCAN_TRACE_CONFIGURE,
                pcan_path_argument,
            )
            log.internal_info("Tracefile configured")

            self._pcan_set_value(
                pcan_channel, PCANBasic.PCAN_TRACE_STATUS, PCANBasic.PCAN_PARAMETER_ON
            )
            log.internal_info("Trace activated")
        except RuntimeError:
            log.error(f"Logging for {self.channel} not activated")
        except OSError as e:
            log.error(f"Can not create log folder for PCAN logs: {e}")
            log.error(f"Logging for {self.channel} not activated")

    def _pcan_set_value(self, channel, parameter, buffer) -> None:
        """Set a value in the PCAN api.

        If this is not successful, a RuntimeError is returned, as well as the
        PCAN error text is logged, if possible.

        :param channel: Channel for PCANBasic.SetValue
        :param parameter: Parameter for PCANBasic.SetValue
        :param buffer: Buffer for PCANBasic.SetValue

        :raises RuntimeError: Raised if the function is not successful
        """
        try:
            result = self.raw_pcan_interface.SetValue(
                channel,
                parameter,
                buffer,
            )
        except Exception as e:
            log.error(f"Exception in call to SetValue: {e}")
            raise RuntimeError("Error configuring logging on PCAN")
        else:
            if result != PCANBasic.PCAN_ERROR_OK:
                _, error_msg = self.raw_pcan_interface.GetErrorText(result)
                log.error(error_msg)
                raise RuntimeError(f"Error configuring logging on PCAN: {result}")

    def _cc_close(self) -> None:
        """Close the current can bus channel and uninitialize PCAN handle."""
        self.bus.shutdown()
        self.bus = None
        if self.logging_activated:
            try:
                result = self.raw_pcan_interface.Uninitialize(PCANBasic.PCAN_NONEBUS)
            except Exception as e:
                log.exception(f"Error in call to Uninitialize: {e}")
            else:
                if result != PCANBasic.PCAN_ERROR_OK:
                    _, error_msg = self.raw_pcan_interface.GetErrorText(result)
                    log.error(error_msg)
            finally:
                self.raw_pcan_interface = None

    def _cc_send(
        self, msg: MessageType, remote_id: Optional[int] = None, **kwargs
    ) -> None:
        """Send a CAN message at the configured id.

        If remote_id parameter is not given take configured ones

        :param msg: data to send
        :param remote_id: destination can id used
        :param kwargs: named arguments
        """

        remote_id = remote_id or self.remote_id

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=msg,
            is_extended_id=self.is_extended_id,
            is_fd=self.is_fd,
            bitrate_switch=self.enable_brs,
        )
        self.bus.send(can_msg)

        log.internal_debug(f"{self} sent CAN Message: {can_msg}, data: {msg}")

    def _cc_receive(
        self, timeout: float = 0.0001
    ) -> Dict[str, Union[bytes, int, None]]:
        """Receive a can message using configured filters.

        :param timeout: timeout applied on reception

        :return: the received data and the source can id
        """
        try:  # Catch bus errors & rcv.data errors when no messages where received
            received_msg = self.bus.recv(timeout=timeout or self.timeout)

            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp

                log.internal_debug(
                    f"received CAN Message: {frame_id}, {payload}, {timestamp}"
                )
                return {"msg": payload, "remote_id": frame_id}
            else:
                return {"msg": None}
        except can.CanError as can_error:
            log.internal_debug(f"encountered can error: {can_error}")
            return {"msg": None}
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Test Execution
**************

:module: test_execution

:synopsis: Execute a test environment based on the supplied configuration.

.. currentmodule:: test_execute

.. note::
    1. Glob a list of test-suite folders
    2. Generate a list of test-suites with a list of test-cases
    3. Loop per suite
    4. Gather result
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .test_case import BasicTest

import enum
import logging
import re
import time
import unittest
from collections import OrderedDict, namedtuple
from fnmatch import fnmatch
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import xmlrunner

import pykiso

from ..exceptions import AuxiliaryCreationError, TestCollectionError
from ..logging_initializer import get_logging_options
from ..test_result.assert_step_report import (
    StepReportData,
    assert_decorator,
    generate_step_report,
)
from ..test_result.text_result import BannerTestResult, ResultStream
from ..test_result.xml_result import XmlTestResult
from . import test_suite

log = logging.getLogger(__name__)

TestFilterPattern = namedtuple("TestFilterPattern", "test_file, test_class, test_case")


@enum.unique
class ExitCode(enum.IntEnum):
    """List of possible exit codes"""

    ALL_TESTS_SUCCEEDED = 0
    ONE_OR_MORE_TESTS_FAILED = 1
    ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION = 2
    ONE_OR_MORE_TESTS_FAILED_AND_RAISED_UNEXPECTED_EXCEPTION = 3
    AUXILIARY_CREATION_FAILED = 4
    BAD_CLI_USAGE = 5


def create_test_suite(
    test_suite_dict: Dict[str, Union[str, int]]
) -> test_suite.BasicTestSuite:
    """create a test suite based on the config dict

    :param test_suite_dict: dict created from config with keys 'suite_dir',
        'test_filter_pattern', 'test_suite_id'
    """
    return test_suite.BasicTestSuite(
        test_suite_dict["suite_dir"],
        test_suite_dict["test_filter_pattern"],
        test_suite_dict["test_suite_id"],
    )


def apply_tag_filter(
    all_tests_to_run: unittest.TestSuite, usr_tags: Dict[str, List[str]]
) -> None:
    """Filter the test cases based on user tags provided via CLI.

    :param all_tests_to_run: a dict containing all testsuites and testcases
    :param usr_tags: encapsulate user's variant choices
    """

    def format_tag_names(tags: Dict[str, List[str]]) -> Dict[str, List[str]]:
        """Remove any comma or underscore from the provided dict's keys."""
        for char in ["_", "-"]:
            tags = {name.replace(char, ""): value for name, value in tags.items()}
        return tags

    def should_skip(test_case: BasicTest) -> bool:
        """Check if test shall be skipped by evaluating the test case tag
        attribute.

        :param test_case: test_case to check
        :return: True if test shall be skipped else False
        """
        if test_case.tag is None:
            return False

        # Skip only the test cases that have a matching tag name but no matching tag value
        for cli_tag_id, cli_tag_value in usr_tags.items():
            # skip any test case that doesn't define a CLI-provided tag name
            if cli_tag_id not in test_case.tag.keys():
                test_case._skip_msg = (
                    f"provided tag {cli_tag_id!r} not present in test tags"
                )
                return True
            # skip any test case that which tag value don't match the provided tag's value
            cli_tag_values = (
                cli_tag_value if isinstance(cli_tag_value, list) else [cli_tag_value]
            )
            if not any(
                cli_val in test_case.tag[cli_tag_id] for cli_val in cli_tag_values
            ):
                test_case._skip_msg = f"non-matching value for tag {cli_tag_id!r}"
                return True

        return False

    def set_skipped(test_case: BasicTest) -> BasicTest:
        """Set testcase to skipped.

        :param test_case: testcase to be skipped
        :return: the skipped testcase as a new instance of the provided
            TestCase subclass.
        """
        skipped_test_cls = unittest.skip(test_case._skip_msg)(test_case.__class__)
        return skipped_test_cls()

    # collect and reformat all CLI and test case tag names
    usr_tags = format_tag_names(usr_tags)

    all_test_tags = []
    for tc in test_suite.flatten(all_tests_to_run):
        if getattr(tc, "tag", None) is None:
            continue
        tc.tag = format_tag_names(tc.tag)
        all_test_tags.extend(tc.tag.keys())

    # verify that each provided tag name is defined in at least one test case
    all_test_tags = set(all_test_tags)
    for tag_name in usr_tags:
        if tag_name not in all_test_tags:
            raise NameError(
                f"Provided tag {tag_name!r} is not defined in any testcase."
            )

    # skip the tests according to the provided CLI tags and the defined test tags
    list(map(set_skipped, filter(should_skip, test_suite.flatten(all_tests_to_run))))


def apply_test_case_filter(
    all_tests_to_run: unittest.TestSuite,
    test_class_pattern: str,
    test_case_pattern: str,
) -> unittest.TestSuite:
    """Apply a filter to run only test cases which matches given expression

    :param all_tests_to_run: a dict containing all testsuites and testcases
    :param test_class_pattern: pattern to select test class as unix filename pattern
    :param test_case_pattern: pattern to select test case as unix filename pattern
    :return: new test suite with filtered test cases
    """

    def is_active_test(test_case: BasicTest) -> bool:
        """Check if testcase shall be active by given selection patterns

        :param test_case: unittest test
        :return: True if test matches patterns else False
        """
        test_class_name = re.sub(r"-\d+-\d+", "", test_case.__class__.__name__)
        is_classname_match = bool(fnmatch(test_class_name, test_class_pattern))

        if is_classname_match and test_case_pattern is None:
            return is_classname_match
        elif is_classname_match and test_case_pattern:
            return bool(fnmatch(test_case._testMethodName, test_case_pattern))
        else:
            return False

    base_suite = test_suite.flatten(all_tests_to_run)
    filtered_suite = filter(is_active_test, base_suite)
    return unittest.TestSuite(filtered_suite)


def failure_and_error_handling(result: unittest.TestResult) -> int:
    """provide necessary information to Jenkins if an error occur during tests execution

    :param result: encapsulate all test results from the current run

    :return: an ExitCode object
    """
    failed, errored = result.failures, result.errors
    if failed and errored:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_FAILED_AND_RAISED_UNEXPECTED_EXCEPTION
    elif failed:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_FAILED
    elif errored:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    else:
        exit_code = ExitCode.ALL_TESTS_SUCCEEDED
    return exit_code


def enable_step_report(
    all_tests_to_run: unittest.suite.TestSuite, step_report: Path
) -> None:
    """Decorate all assert method from Test-Case.

    This will allow to save the assert inputs in
    order to generate the HTML step report.

    :param all_tests_to_run: a dict containing all testsuites and testcases
    """

    # Step report header fed during test
    base_suite = test_suite.flatten(all_tests_to_run)
    for tc in base_suite:
        tc.step_report = StepReportData()
        if step_report is not None:
            # for any test, show ITF version
            tc.step_report.header = OrderedDict({"ITF version": pykiso.__version__})
            # Decorate All assert method
            assert_method_list = [
                method for method in dir(tc) if method.startswith("assert")
            ]
            for method_name in assert_method_list:
                # Get method from name
                method = getattr(tc, method_name)
                # Add decorator to the existing method
                setattr(tc, method_name, assert_decorator(method))


def parse_test_selection_pattern(pattern: str) -> TestFilterPattern:
    """Parse test selection pattern from cli.

    For example: ``test_file.py::TestCaseClass::test_method``

    :param pattern: test selection pattern
    :return: pattern for file, class name and test case name
    """
    if not pattern:
        return TestFilterPattern(None, None, None)
    parsed_patterns = []
    patterns = pattern.split("::")
    for pattern in patterns:
        if pattern == "":
            parsed_patterns.append(None)
        else:
            parsed_patterns.append(pattern)
    for _ in range(3 - (len(parsed_patterns))):
        parsed_patterns.append(None)

    return TestFilterPattern(*parsed_patterns)


def collect_test_suites(
    config_test_suite_list: List[Dict[str, Union[str, int]]],
    test_filter_pattern: Optional[str] = None,
) -> List[Optional[test_suite.BasicTestSuite]]:
    """Collect and load all test suites defined in the test configuration.

    :param config_test_suite_list: list of dictionaries from the configuration
        file corresponding each to one test suite.
    :param test_filter_pattern: optional filter pattern to overwrite
        the one defined in the test suite configuration.

    :raises pykiso.TestCollectionError: if any test case inside one of
        the configured test suites failed to be loaded.

    :return: a list of all loaded test suites.
    """
    list_of_test_suites = []
    for test_suite_configuration in config_test_suite_list:
        try:
            if test_filter_pattern is not None:
                test_suite_configuration["test_filter_pattern"] = test_filter_pattern
            current_test_suite = create_test_suite(test_suite_configuration)
            list_of_test_suites.append(current_test_suite)
        except BaseException as e:
            raise TestCollectionError(test_suite_configuration["suite_dir"]) from e
    return list_of_test_suites


def execute(
    config: Dict[str, Any],
    report_type: str = "text",
    user_tags: Optional[Dict[str, List[str]]] = None,
    step_report: Optional[Path] = None,
    pattern_inject: Optional[str] = None,
    failfast: bool = False,
) -> int:
    """Create test environment based on test configuration.

    :param config: dict from converted YAML config file
    :param report_type: str to set the type of report wanted, i.e. test
        or junit
    :param user_tags: test case tags to execute
    :param step_report: file path for the step report or None
    :param pattern_inject: optional pattern that will override
        test_filter_pattern for all suites. Used in test development to
        run specific tests.
    :param failfast: stop the test run on the first error or failure.

    :return: exit code corresponding to the result of the test execution
        (tests failed, unexpected exception, ...)
    """
    try:
        test_file_pattern = parse_test_selection_pattern(pattern_inject)

        test_suites = collect_test_suites(
            config["test_suite_list"], test_file_pattern.test_file
        )
        # Group all the collected test suites in one global test suite
        all_tests_to_run = unittest.TestSuite(test_suites)

        # filter test cases based on variant and branch-level options
        if user_tags:
            apply_tag_filter(all_tests_to_run, user_tags)

        # Enable step report
        enable_step_report(all_tests_to_run, step_report)

        if test_file_pattern.test_class:
            all_tests_to_run = apply_test_case_filter(
                all_tests_to_run,
                test_file_pattern.test_class,
                test_file_pattern.test_case,
            )

        log_file_path = get_logging_options().log_path
        # TestRunner selection: generate or not a junit report. Start the tests and publish the results
        if report_type == "junit":
            junit_report_name = time.strftime("TEST-pykiso-%Y-%m-%d_%H-%M-%S.xml")
            project_folder = Path.cwd()
            reports_path = project_folder / "reports"
            junit_report_path = reports_path / junit_report_name
            reports_path.mkdir(exist_ok=True)
            with open(junit_report_path, "wb") as junit_output, ResultStream(
                log_file_path
            ) as stream:
                test_runner = xmlrunner.XMLTestRunner(
                    output=junit_output,
                    resultclass=XmlTestResult,
                    failfast=failfast,
                    verbosity=0,
                    stream=stream,
                )
                result = test_runner.run(all_tests_to_run)
        else:
            with ResultStream(log_file_path) as stream:
                test_runner = unittest.TextTestRunner(
                    stream=stream,
                    resultclass=BannerTestResult,
                    failfast=failfast,
                    verbosity=0,
                )
                result = test_runner.run(all_tests_to_run)

        # Generate the html step report
        if step_report is not None:
            generate_step_report(result, step_report)

        exit_code = failure_and_error_handling(result)
    except NameError:
        log.exception("Error occurred during tag evaluation.")
        exit_code = ExitCode.BAD_CLI_USAGE
    except TestCollectionError:
        log.exception("Error occurred during test collection.")
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    except AuxiliaryCreationError:
        log.exception("Error occurred during auxiliary creation.")
        exit_code = ExitCode.AUXILIARY_CREATION_FAILED
    except KeyboardInterrupt:
        log.exception("Keyboard Interrupt detected")
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    except Exception:
        log.exception(f'Issue detected in the test-suite: {config["test_suite_list"]}!')
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    return int(exit_code)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
uds_auxiliary
*************

:module: uds_auxiliary

:synopsis: Auxiliary used to handle Unified Diagnostic Service protocol

.. currentmodule:: uds_auxiliary

"""
import logging
import threading
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator, List, Optional, Union

import can
from uds import IsoServices

from pykiso.connector import CChannel

from .common import uds_exceptions
from .common.uds_base_auxiliary import UdsBaseAuxiliary
from .common.uds_request import UDSCommands
from .common.uds_response import UdsResponse
from .common.uds_utils import get_uds_service

log = logging.getLogger(__name__)


class UdsAuxiliary(UdsBaseAuxiliary):
    """Auxiliary used to handle the UDS protocol on client (tester) side."""

    errors = uds_exceptions

    def __init__(
        self,
        com: CChannel,
        config_ini_path: Optional[Union[Path, str]] = None,
        odx_file_path: Optional[Union[Path, str]] = None,
        request_id: Optional[int] = None,
        response_id: Optional[int] = None,
        tp_layer: dict = None,
        uds_layer: dict = None,
        **kwargs,
    ):
        """Initialize attributes.
        :param com: communication channel connector.
        :param config_ini_path: UDS parameter file.
        :param odx_file_path: ecu diagnostic definition file.
        :param request_id: optional CAN ID used for sending messages.
        :param response_id: optional CAN ID used for receiving messages.
        :param tp_layer: isotp configuration given at yaml level
        :param uds_layer: uds configuration given at yaml level
        """
        self.is_tester_present = None
        super().__init__(
            com,
            config_ini_path,
            odx_file_path,
            request_id,
            response_id,
            tp_layer,
            uds_layer,
            **kwargs,
        )

    def transmit(self, data: bytes, req_id: int, extended: bool = False) -> None:
        """Transmit a message through ITF connector. This method is a
        substitute to transmit method present in python-uds package.

        :param data: data to send
        :param req_id: CAN message identifier
        :param extended: True if addressing mode is extended otherwise
            False
        """
        self.channel._cc_send(msg=data, remote_id=req_id)

    def send_uds_raw(
        self,
        msg_to_send: Union[bytes, List[int], tuple],
        timeout_in_s: float = 6,
        response_required: bool = True,
    ) -> Union[UdsResponse, bool]:
        """Send a UDS diagnostic request to the target ECU and check response.

        :param msg_to_send: can uds raw bytes to be sent
        :param timeout_in_s: not used, actual timeout in seconds for the response can be
            configured with the P2_CAN_Client parameter in the config.ini file
            (default value is 5s)
        :param response_required: Wait for a response if True

        :raise ResponseNotReceivedError: raised when no answer has been received
        :raise Exception: raised when the raw message could not be send properly

        :return: the raw uds response's bytes, or True if a response is
            not expected and the command is properly sent otherwise
            False
        """
        try:
            log.internal_info(
                f"UDS request to send '{['0x{:02X}'.format(i) for i in msg_to_send]}'"
            )
            resp = self.uds_config.send(
                msg_to_send,
                responseRequired=response_required,
                tpWaitTime=self.tp_waiting_time,
            )
        except Exception:
            log.exception("Error while sending uds raw request")
            return False

        if resp is None:
            if not response_required:
                return True
            else:
                raise self.errors.ResponseNotReceivedError(msg_to_send)

        resp_print = (
            f"UDS response received {['0x{:02X}'.format(i) for i in resp]}"
            if not isinstance(resp, bool)
            else resp
        )
        log.internal_info(resp_print)
        resp = UdsResponse(resp)
        return resp

    def check_raw_response_positive(self, resp: UdsResponse) -> Optional[bool]:
        """Check if the response is positive, raise an error if not

        :param resp: raw response of uds request

        :raise UnexpectedResponseError: raised when the answer is not the expected one

        :return: True if response is positive
        """
        if resp.is_negative:
            log.internal_info(f"Negative response with NRC: {resp.nrc.name}")
            raise self.errors.UnexpectedResponseError(resp)
        return True

    def check_raw_response_negative(self, resp: UdsResponse) -> Optional[bool]:
        """Check if the response is negative, raise an error if not

        :param resp: raw response of uds request

        :raise UnexpectedResponseError: raised when the answer is not the expected one

        :return: True if response is negative
        """
        if not resp.is_negative:
            raise self.errors.UnexpectedResponseError(resp)
        log.internal_info(f"Negative response with :{resp.nrc.name}")
        return True

    def send_uds_config(
        self,
        msg_to_send: dict,
        timeout_in_s: float = 6,
    ) -> Union[dict, bool]:
        """Send UDS config to the target ECU.

        :param msg_to_send: uds config to be sent
        :param timeout_in_s: not used

        :return: a dict containing the uds response, or True if a
            response is not expected and the command is properly sent
            otherwise False
        """
        if not self.uds_config_enable:
            log.error("Uds configured without ODX, sending Uds config impossible!")
            return False

        try:
            uds_service_name = get_uds_service(msg_to_send["service"])
            uds_service = getattr(self.uds_config, uds_service_name)
            req_resp_data = uds_service(**msg_to_send["data"])
            if req_resp_data is None:
                req_resp_data = True
            log.internal_info(f"UDS response received {req_resp_data}")
            return req_resp_data
        except AttributeError:
            # Service not found, raised by getattr()
            log.exception(
                f"Could not send UDS config request: unknown service {uds_service_name}"
            )
            return False
        except Exception:
            # Data send failed
            log.exception("Could not send UDS config request: sending failed")
            return False

    def hard_reset(self) -> Union[dict, UdsResponse]:
        """Allow power reset of the component

        :return: response of the hard reset request
        """
        return self.send_uds_raw(UDSCommands.ECUReset.HARD_RESET)

    def force_ecu_reset(self) -> UdsResponse:
        """Allow power reset of the component

        :return: response of the force ecu reset request
        """
        if not self.uds_config_enable:
            return self.send_uds_raw(UDSCommands.ECUReset.KEY_OFF_ON_RESET)
        else:
            ecu_hard_reset_req = {
                "service": IsoServices.EcuReset,
                "data": {"parameter": "ForceEcuReset"},
            }
            return self.send_uds_config(ecu_hard_reset_req)

    def soft_reset(self) -> Union[dict, UdsResponse]:
        """Perform soft reset of the component, equivalent to a restart of application

        :return: response of the soft reset request
        """
        if not self.uds_config_enable:
            return self.send_uds_raw(UDSCommands.ECUReset.SOFT_RESET)
        else:
            ecu_soft_reset_req = {
                "service": IsoServices.EcuReset,
                "data": {"parameter": "Soft Reset"},
            }
            return self.send_uds_config(ecu_soft_reset_req)

    def read_data(self, parameter: str) -> Optional[Union[dict, bool]]:
        """UDS config command that allow data reading

        :param parameter: data to be read

        :return: a dict with uds config response
        """
        if self.uds_config_enable:
            req = {
                "service": IsoServices.ReadDataByIdentifier,
                "data": {
                    "parameter": parameter,
                },
            }
            return self.send_uds_config(req)
        else:
            log.error("No uds config found")
            return

    def write_data(
        self, parameter: str, value: Union[List[bytes], bytes]
    ) -> Optional[Union[dict, bool]]:
        """UDS config command that allow data writing

        :param parameter: data to be set
        :param value: new content of the data

        :return: a dict with uds config response
        """
        if self.uds_config_enable:
            req = {
                "service": IsoServices.WriteDataByIdentifier,
                "data": {"parameter": parameter, "dataRecord": value},
            }
            return self.send_uds_config(req)
        else:
            log.error("No uds config found")
            return

    def _sender_run(self, period: int, stop_event: threading.Event) -> None:
        """send tester present at defined period until stopped

        :param period: period in seconds to use for the cyclic sending of tester present
        :param stop_event: event to set to stop the sending of tester present
        """
        while not stop_event.is_set():
            self.send_uds_raw(
                UDSCommands.TesterPresent.TESTER_PRESENT_NO_RESPONSE,
                response_required=False,
            )
            time.sleep(period)

    @contextmanager
    def tester_present_sender(self, period: int = 4) -> Iterator[None]:
        """Context manager that continuously sends tester present messages via UDS

        :param period: period in seconds to use for the cyclic sending of tester present
        """
        stop_event = threading.Event()
        sender = threading.Thread(
            name="TesterPresentSender",
            target=self._sender_run,
            args=(period, stop_event),
        )
        try:
            yield sender.start()
        finally:
            stop_event.set()
            sender.join()

    def start_tester_present_sender(self, period: int = 4):
        """Start to continuously send tester present messages via UDS"""
        if not self.is_tester_present:
            self.is_tester_present = self.tester_present_sender(period=period)
            return self.is_tester_present.__enter__()

    def stop_tester_present_sender(self):
        """Stop to continuously send tester present messages via UDS"""
        if self.is_tester_present:
            self.is_tester_present.__exit__(None, None, None)
            self.is_tester_present = None
        else:
            log.internal_warning(
                "Tester present sender should be started before it can be stopped"
            )

    def _receive_message(self, timeout_in_s: float) -> None:
        """This method is only used to populate the python-uds reception
        buffer. When a message is received, invoke python-uds configured
        callback to make it available for pyton-uds

        :param timeout_in_s: timeout on reception.
        """

        recv_response = self.channel.cc_receive(timeout=timeout_in_s)
        received_data = recv_response.get("msg")
        arbitration_id = recv_response.get("remote_id")

        if received_data is not None and arbitration_id == self.res_id:
            can_msg = can.Message(
                data=received_data, arbitration_id=arbitration_id, is_extended_id=False
            )
            self.uds_config.tp.callback_onReceive(can_msg)

    def _run_command(self, cmd_message, cmd_data=None) -> Union[dict, bytes, bool]:
        """Not used."""
        pass

    def _delete_auxiliary_instance(self) -> bool:
        """Close current associated channel.

        :return: always True
        """
        if self.is_tester_present:
            self.stop_tester_present_sender()
        return super()._delete_auxiliary_instance()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pykiso',
 'pykiso.interfaces',
 'pykiso.lib',
 'pykiso.lib.auxiliaries',
 'pykiso.lib.auxiliaries.instrument_control_auxiliary',
 'pykiso.lib.auxiliaries.simulated_auxiliary',
 'pykiso.lib.auxiliaries.udsaux',
 'pykiso.lib.auxiliaries.udsaux.common',
 'pykiso.lib.connectors',
 'pykiso.lib.connectors.cc_socket_can',
 'pykiso.lib.robot_framework',
 'pykiso.test_coordinator',
 'pykiso.test_result',
 'pykiso.test_setup',
 'pykiso.tool',
 'pykiso.tool.pykiso_to_pytest',
 'pykiso.tool.testrail']

package_data = \
{'': ['*'],
 'pykiso.test_result': ['templates/*'],
 'pykiso.tool.pykiso_to_pytest': ['templates/*']}

install_requires = \
['Jinja2>=2.11.0,<4.0.0',
 'MarkupSafe>=2.0.1,<2.1.0',
 'PyVISA-py>=0.5.3,<0.6.0',
 'PyVISA>=1.12.0,<2.0.0',
 'PyYAML>=6.0,<7.0',
 'brainstem',
 'click>=7.0.0,<9.0.0',
 'hidapi>=0.12.0.post2,<0.13.0',
 'pykiso-python-uds>=3.0.2,<3.1.0',
 'pylink-square>=0.12,<0.15',
 'pyserial>=3.0,<4.0',
 'python-can[pcan]>=4.1.0,<4.2.0',
 'robotframework==3.2.2',
 'tabulate>=0.8.9,<0.10.0',
 'unittest-xml-reporting>=3.2.0,<4.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.12,<6.0'],
 ':python_version >= "3.5" and python_version < "4.0"': ['pyreadline3>=3.4.1,<4.0.0'],
 'all': ['rich>=13.2.0,<14.0.0', 'requests>=2.28.2,<3.0.0'],
 'testrail': ['rich>=13.2.0,<14.0.0', 'requests>=2.28.2,<3.0.0']}

entry_points = \
{'console_scripts': ['instrument-control = '
                     'pykiso.lib.auxiliaries.instrument_control_auxiliary.instrument_control_cli:main',
                     'pykiso = pykiso.cli:main',
                     'pykiso-tags = pykiso.tool.show_tag:main',
                     'pykitest = pykiso.tool.pykiso_to_pytest.cli:main',
                     'testrail = pykiso.tool.testrail.cli:cli_testrail']}

setup_kwargs = {
    'name': 'pykiso',
    'version': '0.21.0',
    'description': 'Embedded integration testing framework.',
    'long_description': '[![License](https://img.shields.io/badge/Licence-Eclipse%20Public%20License%202.0-lightgrey)](https://opensource.org/licenses/EPL-2.0)\n[![Platforms](https://img.shields.io/badge/Platforms-win64%20linux64%20osx64-lightgrey)]()\n[![Supported python version](https://img.shields.io/pypi/pyversions/pykiso)]()\n[![Build status](https://img.shields.io/jenkins/build?jobUrl=https%3A%2F%2Fci.eclipse.org%2Fkiso-testing%2Fjob%2Fkiso-testing%2Fjob%2Fmaster%2F)](https://ci.eclipse.org/kiso-testing/job/kiso-testing/job/master/)\n[![Documentation Status](https://readthedocs.org/projects/kiso-testing/badge/?version=latest)](https://kiso-testing.readthedocs.io/en/latest/?badge=latest)\n[![Test results](https://img.shields.io/jenkins/tests?compact_message&failed_label=failed&jobUrl=https%3A%2F%2Fci.eclipse.org%2Fkiso-testing%2Fjob%2Fkiso-testing%2Fjob%2Fmaster%2F&passed_label=passed&skipped_label=skipped)](https://ci.eclipse.org/kiso-testing/job/kiso-testing/job/master/)\n[![codecov](https://codecov.io/gh/eclipse/kiso-testing/branch/master/graph/badge.svg?token=IBKQ700ABS)](https://codecov.io/gh/eclipse/kiso-testing)\n[![Code quality](https://img.shields.io/lgtm/grade/python/github/eclipse/kiso-testing)]()\n[![Last commit](https://img.shields.io/github/last-commit/eclipse/kiso-testing)]()\n[![Commits since latest version](https://img.shields.io/github/commits-since/eclipse/kiso-testing/latest/master)]()\n\n# PyKiso\n\n![Optional Text](./docs/images/pykiso_logo.png)\n\n## Introduction ##\n\n**pykiso** is an integration test framework. With it, it is possible to write\n* Whitebox integration tests directly on my target device\n* Graybox integration tests to make sure the communication-link with my target device is working as expected\n* Blackbox integration tests to make sure my external device interfaces are working as expected\n\nThe project will contain:\n* The core python framework (this repository)\n* Framework plugins that are generic enough to be integrated as "native" (this repository)\n* Additional "testApps" for different targets platforms (e.g. stm32, ...) or languages (C, C++, ...) . It could be pure SW or also HW (other repositories)\n\n## Requirements ##\n\n* Python 3.7+\n* pip/poetry (used to get the rest of the requirements)\n\n## Install ##\n\n```bash\npip install pykiso\n```\n\n[Poetry](https://python-poetry.org/) is more appropriate for developers as it automatically creates virtual environments.\n\n```bash\ncd kiso-testing\npoetry install\npoetry shell\n```\n\n### Pre-Commit\n\nTo improve code-quality, a configuration of [pre-commit](https://pre-commit.com/) hooks are available.\nThe following pre-commit hooks are used:\n\n- black\n- flake8\n- isort\n- trailing-whitespace\n- end-of-file-fixer\n- check-docstring-first\n- check-json\n- check-added-large-files\n- check-yaml\n- debug-statements\n\nIf you don\'t have pre-commit installed, you can get it using pip:\n\n```bash\npip install pre-commit\n```\n\nStart using the hooks with\n\n```bash\npre-commit install\n```\n\n## Commit message convention\n\nCommits are sorted into multiple categories based on keywords that can occur at any position as part of the commit message.\n[Category] Keywords\n* [BREAKING CHANGES] BREAKING CHANGE\n* [Features] feat:\n* [Fixes] fix:\n* [Docs] docs:\n* [Styles] style:\n* [Refactors] refactor!:\n* [Performances] perf:\n* [Tests] test:\n* [Build] build:\n* [Ci] ci:\nEach commit is considered only once according to the order of the categories listed above. Merge commits are ignored.\n\nThe tool commitizen can help you to create commits which follows these standards.\n```bash\n# if not yet installed:\npip install -U commitizen==2.20.4\n# helps you to create a commit:\ncz commit\n# or use equivalent short variant:\ncz c\n```\n\n## Generate Changelog\n\nAfter you installed the dev dependencies from the pipfile you are able to\nautogenerate the Changelog.\n\n```bash\ninvoke changelog\n```\n\n## Usage ##\n\nOnce installed the application is bound to `pykiso`, it can be called with the following arguments:\n\n```bash\nUsage: pykiso [OPTIONS]\n\n  Embedded Integration Test Framework - CLI Entry Point.\n\n  TAG Filters: any additional option to be passed to the test as tag through\n  the pykiso call. Multiple values must be separated with a comma.\n\n  For example: pykiso -c your_config.yaml --branch-level dev,master --variant\n  delta\n\nOptions:\n  -c, --test-configuration-file FILE\n                                  path to the test configuration file (in YAML\n                                  format)  [required]\n  -l, --log-path PATH             path to log-file or folder. If not set will\n                                  log to STDOUT\n  --log-level [DEBUG|INFO|WARNING|ERROR]\n                                  set the verbosity of the logging\n  --junit                         enables the generation of a junit report\n  --text                          default, test results are only displayed in\n                                  the console\n  --step-report PATH              generate the step report at the specified\n                                  path\n  --failfast                      stop the test run on the first error or\n                                  failure\n  -v, --verbose                   activate the internal framework logs\n  -p, --pattern TEXT              test filter pattern, e.g. \'test_suite_1.py\'\n                                  or \'test_*.py\'. Or even more granularly\n                                  \'test_suite_1.py::test_class::test_name\'\n  --version                       Show the version and exit.\n  -h, --help                      Show this message and exit.\n```\n\nSuitable config files are available in the `examples` folder.\n\n### Demo using example config ##\n\n```bash\ninvoke run\n```\n\n### Running the Tests ##\n\n```bash\ninvoke test\n```\n\nor\n\n```bash\npytest\n```\n',
    'author': 'Sebastian Fischer',
    'author_email': 'sebastian.fischer@de.bosch.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/pykiso/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Integration Test Framework
**************************

:module: cli

:synopsis: Entry point to the integration test framework.

.. currentmodule:: cli

"""
import collections
import logging
import pprint
import sys
import time
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Tuple

import click

from pykiso.test_coordinator.test_execution import ExitCode

from . import __version__
from .config_parser import parse_config
from .global_config import Grabber
from .logging_initializer import initialize_logging
from .test_coordinator import test_execution
from .test_setup.config_registry import ConfigRegistry
from .types import PathType


def eval_user_tags(click_context: click.Context) -> Dict[str, List[str]]:
    """Evaluate commandline args for user tags and raise exceptions for invalid
    arguments.

    :param click_context: click context
    :raises click.NoSuchOption: if key doesnt start with "--" or has an invalid
      character like "_"
    :raises click.BadOptionUsage: no value specfied for user tag
    :return: user tags with values
    """
    user_tags = {}
    user_args = click_context.args.copy()
    if not user_args:
        return user_tags
    while user_args:
        try:
            key = user_args.pop(0)
            if not key.startswith("--") or "_" in key:
                correct_key = (
                    f'{"" if key.startswith("--") else "--" }{key.replace("_","-")}'
                )
                raise click.NoSuchOption(option_name=key, possibilities=[correct_key])

            value = user_args.pop(0)
            user_tags[key[2:]] = value.split(",")
        except IndexError:
            raise click.BadOptionUsage(
                option_name=key,
                message=f"No value specified for tag {key}",
                ctx=click_context,
            )
    return user_tags


def check_file_extension(
    click_context: click.Context, param: click.Parameter, paths: Tuple[str]
) -> Tuple[str]:
    """Check the path strings of all given test configuration files for yaml file
    extension and raise an exception if the check fails

    :click_context: click context (click callback requirement)
    :param: click parameter (click callback requirement)
    :paths: the paths of the files to check
    :raises click.BadParameter: if one of the files does not have .yaml or .yml
        as extension
    :return: the checked paths as they were passed
    """
    for path in paths:
        if not path.endswith((".yaml", ".yml")):
            raise click.BadParameter(
                f"Configuration needs to be a .yaml file, but {path} was given"
            )
    return paths


@click.command(
    context_settings={
        "help_option_names": ["-h", "--help"],
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.option(
    "-c",
    "--test-configuration-file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    multiple=True,
    callback=check_file_extension,
    help="path to the test configuration file (in YAML format)",
)
@click.option(
    "-l",
    "--log-path",
    required=False,
    default=None,
    type=click.Path(writable=True),
    help="path to log-file or folder. If not set will log to STDOUT",
)
@click.option(
    "--log-level",
    required=False,
    default="INFO",
    type=click.Choice(
        "DEBUG INFO WARNING ERROR".split(" "),
        case_sensitive=False,
    ),
    help="set the verbosity of the logging",
)
@click.option(
    "--junit",
    "report_type",
    flag_value="junit",
    required=False,
    help="enables the generation of a junit report",
)
@click.option(
    "--text",
    "report_type",
    flag_value="text",
    required=False,
    default=True,
    help="default, test results are only displayed in the console",
)
@click.option(
    "--step-report",
    required=False,
    default=None,
    type=click.Path(writable=True),
    help="generate the step report at the specified path",
)
@click.option(
    "--failfast",
    is_flag=True,
    help="stop the test run on the first error or failure",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    required=False,
    help="activate the internal framework logs",
)
@click.option(
    "-p",
    "--pattern",
    type=click.STRING,
    required=False,
    help="test filter pattern, e.g. 'test_suite_1.py' or 'test_*.py'. Or even more granularly 'test_suite_1.py::TestClass::test_name'",
)
@click.version_option(__version__)
@Grabber.grab_cli_config
@click.pass_context
def main(
    click_context: click.Context,
    test_configuration_file: Tuple[PathType],
    log_path: PathType = None,
    log_level: str = "INFO",
    report_type: str = "text",
    step_report: Optional[PathType] = None,
    pattern: Optional[str] = None,
    failfast: bool = False,
    verbose: bool = False,
):
    """Embedded Integration Test Framework - CLI Entry Point.

    TAG Filters: any additional option to be passed to the test as tag through
    the pykiso call. Multiple values must be separated with a comma.

    For example: pykiso -c your_config.yaml --branch-level dev,master --variant delta

    \f
    :param click_context: click context
    :param test_configuration_file: path to the YAML config file
    :param log_path: path to an existing directory or file to write logs to
    :param log_level: any of DEBUG, INFO, WARNING, ERROR
    :param report_type: if "test", the standard report, if "junit", a junit report is generated
    :param variant: allow the user to execute a subset of tests based on variants
    :param branch_level: allow the user to execute a subset of tests based on branch levels
    :param step_report: file path for the step report or None
    :param pattern: overwrite the pattern from the YAML file for easier test development
    :param failfast: stop the test run on the first error or failure
    :param verbose: activate logging for the whole framework
    """

    for config_file in test_configuration_file:
        # Set the logging
        logger = initialize_logging(log_path, log_level, verbose, report_type)
        # Get YAML configuration
        cfg_dict = parse_config(config_file)
        # Run tests
        logger.debug("cfg_dict:\n{}".format(pprint.pformat(cfg_dict)))

        ConfigRegistry.register_aux_con(cfg_dict)

        user_tags = eval_user_tags(click_context)

        exit_code = test_execution.execute(
            cfg_dict, report_type, user_tags, step_report, pattern, failfast
        )
        ConfigRegistry.delete_aux_con()
        for handler in logging.getLogger().handlers:
            if isinstance(handler, logging.FileHandler):
                logging.getLogger().removeHandler(handler)

    sys.exit(exit_code)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Double Threaded based Auxiliary Interface
*****************************************

:module: dt_auxiliary

:synopsis: common double threaded based auxiliary interface

.. currentmodule:: dt_auxiliary

"""
import abc
import enum
import functools
import logging
import queue
import threading
from enum import Enum, unique
from typing import Any, Callable, List, Optional

from ..exceptions import AuxiliaryCreationError
from ..logging_initializer import add_internal_log_levels, initialize_loggers

log = logging.getLogger(__name__)


@unique
class AuxCommand(Enum):
    """Contain all available auxiliary's commands."""

    #: create auxiliary command id
    CREATE_AUXILIARY = enum.auto()
    #: delete auxiliary command id
    DELETE_AUXILIARY = enum.auto()


class DTAuxiliaryInterface(abc.ABC):
    """Common interface for all double threaded auxiliary. A so called
    << double threaded >> auxiliary, simply encapsulate two threads one
    for the reception and one for the transmmission.
    """

    def __init__(
        self,
        name: str = None,
        is_proxy_capable: bool = False,
        connector_required: bool = True,
        activate_log: List[str] = None,
        tx_task_on=True,
        rx_task_on=True,
        auto_start: bool = True,
    ) -> None:
        """Initialize auxiliary attributes

        :param name: alias of the auxiliary instance
        :param is_proxy_capable: notify if the current auxiliary could
            be (or not) associated to a proxy-auxiliary.
        :param connector_required: define if a connector is required for
            this auxiliary.
        :param activate_log: loggers to deactivate
        :param tx_task_on: enable or not the tx thread
        :param rx_task_on: enable or not the rx thread
        :param auto_start: determine if the auxiliayry is automatically
             started (magic import) or manually (by user)
        """
        initialize_loggers(activate_log)
        add_internal_log_levels()
        self.name = name
        self.is_proxy_capable = is_proxy_capable
        self.auto_start = auto_start
        self.lock = threading.RLock()
        self.stop_tx = threading.Event()
        self.stop_rx = threading.Event()
        self.queue_in = queue.Queue()
        self.queue_out = queue.Queue()
        self.tx_task_on = tx_task_on
        self.rx_task_on = rx_task_on
        self.tx_thread = None
        self.rx_thread = None
        self.recv_timeout = 1
        self.is_instance = False
        self.connector_required = connector_required

    def run_command(
        self,
        cmd_message: Any,
        cmd_data: Any = None,
        blocking: bool = True,
        timeout_in_s: int = 5,
    ) -> Any:
        """Send a request by transmitting it through queue_in and
        waiting for a response using queue_out.

        :param cmd_message: command request to the auxiliary
        :param cmd_data: data you would like to populate the command
            with
        :param blocking: If you want the command request to be
            blocking or not
        :param timeout_in_s: Number of time (in s) you want to wait
            for an answer

        :return: True if the request is correctly executed otherwise
            False
        """
        with self.lock:
            log.internal_debug(
                f"sending command '{cmd_message}' with payload {cmd_data} using {self.name} aux."
            )
            response_received = None
            self.queue_in.put((cmd_message, cmd_data))
            try:
                response_received = self.queue_out.get(blocking, timeout_in_s)
                log.internal_debug(
                    f"reply to command '{cmd_message}' received: '{response_received}' in {self.name}"
                )
            except queue.Empty:
                log.error(
                    f"no reply received within time for command {cmd_message} for payload {cmd_data} using {self.name} aux."
                )
        return response_received

    def create_instance(self) -> bool:
        """Start auxiliary's running tasks and activities.

        :return: True if the auxiliary is created otherwise False

        :raises AuxiliaryCreationError: if instance creation failed
        """
        log.internal_info(f"Creating instance of auxiliary {self.name}")

        with self.lock:
            # if the current aux is alive don't try to create it again
            if self.is_instance:
                return True

            is_created = self._create_auxiliary_instance()

            if not is_created:
                raise AuxiliaryCreationError(self.name)

            # start each auxiliary's tasks
            self._start_tx_task()
            self._start_rx_task()

            self.is_instance = True
            return is_created

    def delete_instance(self) -> bool:
        """Stop auxiliary's running tasks and activities.

        :return: True if the auxiliary is deleted otherwise False
        """
        log.internal_info(f"Deleting instance of auxiliary {self.name}")

        with self.lock:

            # if the current aux is not alive don't try to delete it
            # again
            if not self.is_instance:
                return True

            # stop each auxiliary's tasks
            self._stop_tx_task()
            self._stop_rx_task()

            is_deleted = self._delete_auxiliary_instance()

            if not is_deleted:
                log.error(
                    f"Unexpected error occurred during deletion of auxiliary instance {self.name}"
                )

            self.is_instance = False
            return is_deleted

    def _start_tx_task(self) -> None:
        """Start transmission task."""
        if self.tx_task_on is False:
            log.internal_debug("transmit task is not needed, don't start it")
            return

        log.internal_debug(f"start transmit task {self.name}_tx")
        self.tx_thread = threading.Thread(
            name=f"{self.name}_tx", target=self._transmit_task
        )
        self.tx_thread.start()

    def _start_rx_task(self) -> None:
        """Start reception task."""
        if self.rx_task_on is False:
            log.internal_debug("reception task is not needed, don't start it")
            return

        log.internal_debug(f"start reception task {self.name}_rx")
        self.rx_thread = threading.Thread(
            name=f"{self.name}_rx", target=self._reception_task
        )
        self.rx_thread.start()

    def _stop_tx_task(self) -> None:
        """Stop transmission task."""
        if self.tx_task_on is False:
            log.internal_debug("transmit task was not started, so no need to stop it")
            return

        log.internal_debug(f"stop transmit task {self.name}_tx")
        self.queue_in.put((AuxCommand.DELETE_AUXILIARY, None))
        self.stop_tx.set()
        self.tx_thread.join()
        self.stop_tx.clear()

    def _stop_rx_task(self) -> None:
        """Stop reception task."""
        if self.rx_task_on is False:
            log.internal_debug("recpetion task was not started, so no need t stop it")
            return

        log.internal_debug(f"stop reception task {self.name}_rx")
        self.stop_rx.set()
        self.rx_thread.join()
        self.stop_rx.clear()

    def start(self) -> bool:
        """Force the auxiliary to start all running tasks and
        activities.

        .. warning:: due to the usage of create_instance if an issue
            occurred the exception AuxiliaryCreationError is raised.

        :return: True if the auxiliary is started otherwise False
        """

        return self.create_instance()

    def stop(self) -> bool:
        """Force the auxiliary to stop all running tasks and activities.

        :return: True if the auxiliary is stopped otherwise False
        """
        return self.delete_instance()

    def suspend(self) -> bool:
        """Supend current auxiliary's run.

        :return: True if the auxiliary is suspend otherwise False
        """
        return self.delete_instance()

    def resume(self) -> bool:
        """Resume current auxiliary's run.

        .. warning:: due to the usage of create_instance if an issue
            occurred the exception AuxiliaryCreationError is raised.

        :return: True if the auxiliary is resumed otherwise False
        """
        return self.create_instance()

    def _transmit_task(self) -> None:
        """Auxiliary transmission task.

        Simply call the child defined _run_command.
        """
        while not self.stop_tx.is_set():
            cmd, data = self.queue_in.get()
            # just stop the current Tx thread task
            if cmd == AuxCommand.DELETE_AUXILIARY:
                break

            self._run_command(cmd, data)

    def _reception_task(self) -> None:
        """Auxiliary reception task.

        Simply call the child defined _receive_message method.
        """
        while not self.stop_rx.is_set():
            self._receive_message(timeout_in_s=self.recv_timeout)

    def wait_for_queue_out(
        self, blocking: bool = False, timeout_in_s: int = 0
    ) -> Optional[Any]:
        """Wait for data from the queue out.

        :param blocking: True: wait for timeout to expire, False: return
            immediately
        :param timeout_in_s: if blocking, wait the defined time in
            seconds

        :return: data contained in the auxiliary's queue_out
        """
        try:
            return self.queue_out.get(blocking, timeout_in_s)
        except queue.Empty:
            return None

    @abc.abstractmethod
    def _create_auxiliary_instance(self) -> bool:
        """Common interface call at auxiliary creation.

        This method could be used to e.g initiate the communication
        using the attached connector.

        :return: True - Successfully created / False - Failed by creation
        """

    @abc.abstractmethod
    def _delete_auxiliary_instance(self) -> bool:
        """Common interface call at auxiliary deletion.

        This method could be used to e.g terminate the current running
        communication...

        :return: True - Successfully deleted / False - Failed deleting
        """

    @abc.abstractmethod
    def _run_command(self, cmd_message: Any, cmd_data: Optional[bytes]) -> None:
        """Run a command for the auxiliary.

        :param cmd_message: command to send
        :param cmd_data: payload data for the command
        """

    @abc.abstractmethod
    def _receive_message(self, timeout_in_s: float) -> None:
        """Defines what needs to be done as a receive message. Such as,
         what do I need to do to receive a message.

        :param timeout_in_s: How much time to block on the receive
        """


def open_connector(func: Callable) -> Callable:
    """Open current associated auxiliary's channel.

    :param func: decorated method

    :return: inner decorated function
    """

    @functools.wraps(func)
    def inner_open(self, *arg, **kwargs) -> bool:
        """Open the channel.

        :param args: positional arguments
        :param kwargs: named arguments

        :return: True if everything was successful otherwise False
        """
        log.internal_info("Open channel")
        try:
            self.channel.open()
            return func(self, *arg, **kwargs)
        except Exception:
            log.exception("Unable to open channel communication")
            return False

    return inner_open


def close_connector(func: Callable) -> Callable:
    """Close current associated auxiliary's channel.

    :param func: decorated method

    :return: inner decorated function
    """

    @functools.wraps(func)
    def inner_close(self, *arg, **kwargs) -> bool:
        """Close the channel.

        :param args: positional arguments
        :param kwargs: named arguments

        :return: True if everything was successful otherwise False
        """
        log.internal_info("Close channel")
        try:
            self.channel.close()
            return func(self, *arg, **kwargs)
        except Exception:
            log.exception("Unable to close channel communication")
            return False

    return inner_close


def flash_target(func: Callable) -> Callable:
    """Flash firmware on the target, using associated auxiliary's
    flasher channel.

    :param func: decorated method

    :return: inner decorated function
    """

    @functools.wraps(func)
    def inner_flash(self, *arg, **kwargs) -> bool:
        """Flash the device under test.

        :param args: positional arguments
        :param kwargs: named arguments

        :return: True if everything was successful otherwise False
        """
        if self.flash is None and not self.is_instance:
            log.internal_debug("No flasher configured!")
            return func(self, *arg, **kwargs)

        try:
            log.internal_info("Flash target")
            with self.flash as flasher:
                flasher.flash()
            return func(self, *arg, **kwargs)
        except Exception:
            log.exception("Error occurred during flashing")
            return False

    return inner_flash
